//############################################################################
//
// ilda.cpp
//
// ~ International Laser Display Association ~
// ~     ILDA File Format Driver in C++      ~
// ~             Built for EZFB              ~
//
// by James Lehman
// james@akrobiz.com
//
// began: October 2003
//
// This software is distributed in the hope that it will be useful, but
// without any warranty; without even the implied warranty of merchantability
// or fitness for a particular purpose. See the GNU General Public License
// for more details.
//
//############################################################################

#include "ilda.hpp"
    
//############################################################################
double pi         = (4 * atan(1.0)) ;
double two_pi     = (pi * 2.0)      ;
double half_pi    = (pi / 2.0)      ;
double one_degree = (pi / 180.0)    ;

//############################################################################
void print_ilda_return(int e)
{
    if(!e)
        cout << "    ILDA_OK" << endl;
    else
    {
        if(e & ILDA_BAD_HEADER           )  cout << "    ILDA_BAD_HEADER"           << endl;
        if(e & ILDA_EOF_ERROR            )  cout << "    ILDA_EOF_ERROR"            << endl;
        if(e & ILDA_FILE_OPEN_FAILED     )  cout << "    ILDA_FILE_OPEN_FAILED"     << endl;
        if(e & ILDA_UNKNOWN_FORMAT       )  cout << "    ILDA_UNKNOWN_FORMAT"       << endl;
        if(e & ILDA_INCOMPATIBLE_FORMATS )  cout << "    ILDA_INCOMPATIBLE_FORMATS" << endl;
    }
    return;
}

//############################################################################
// class ilda_section
//############################################################################
ilda_section::ilda_section(const ilda_section& section)
                           : format    (section.format  ),
                             quantity  (section.quantity),
                             identity  (section.identity),
                             total     (section.total   ),
                             scanner   (section.scanner ),
                             future    (0               ),
                             x         (NULL            ),
                             y         (NULL            ),
                             z         (NULL            ), 
                             k         (NULL            ),
                             c         (NULL            ),
                             r         (NULL            ),
                             g         (NULL            ),
                             b         (NULL            ),
                             error     (ILDA_OK         )
{
    get_mem();
    switch(format)
    {
        case  0: 
        case  1: 
                 memcpy((void*)x, (const void*)section.x, quantity * 2);
                 memcpy((void*)y, (const void*)section.y, quantity * 2);
                 memcpy((void*)z, (const void*)section.z, quantity * 2);
                 memcpy((void*)k, (const void*)section.k, quantity    );
                 memcpy((void*)c, (const void*)section.c, quantity    );
                 break;
        //--------------------------------------------------------------------
        case  2: 
                 memcpy((void*)r, (const void*)section.r, quantity    );
                 memcpy((void*)g, (const void*)section.g, quantity    );
                 memcpy((void*)b, (const void*)section.b, quantity    );
                 break;
        //--------------------------------------------------------------------
        default: error = ILDA_UNKNOWN_FORMAT;
                 break;
    }
}

//############################################################################
ilda_section::ilda_section(ilda_3D_point(*F)(int, int), int points, int frame)
                          :format    (0      ),
                           quantity  (points ),
                           identity  (0      ),
                           total     (0      ),
                           scanner   (0      ),
                           future    (0      ),
                           x         (NULL   ),
                           y         (NULL   ),
                           z         (NULL   ),
                           k         (NULL   ),
                           c         (NULL   ),
                           r         (NULL   ),
                           g         (NULL   ),
                           b         (NULL   ),
                           error     (ILDA_OK)
{
    get_mem();
    for(u_int i = 0; i < quantity; i++)
    {
        x[i] = F(i, frame).x;
        y[i] = F(i, frame).y;
        z[i] = F(i, frame).z;
        k[i] = F(i, frame).k;
        c[i] = F(i, frame).c;
    }
}

//############################################################################
void ilda_section::get_mem()
{
    delete_mem();
    if(quantity)
    {
        switch(format)
        {
            case  0:
            case  1:
                     x = new short[quantity];
                     y = new short[quantity];
                     z = new short[quantity];
                     k = new char [quantity];
                     c = new char [quantity];
                     memset((void*)x, 0, quantity * 2);
                     memset((void*)y, 0, quantity * 2);
                     memset((void*)z, 0, quantity * 2);
                     memset((void*)k, 0, quantity    );
                     memset((void*)c, 0, quantity    );
                     break;
            //----------------------------------------------------------------
            case  2: r = new char [quantity];
                     g = new char [quantity];
                     b = new char [quantity];
                     memset((void*)r, 0, quantity    );
                     memset((void*)g, 0, quantity    );
                     memset((void*)b, 0, quantity    );
                     break;
            //----------------------------------------------------------------
            default: error = ILDA_UNKNOWN_FORMAT;
                     break;
        }
    }
    return;
}

//############################################################################
void ilda_section::delete_mem()
{
    if(x) delete x;
    if(y) delete y;
    if(z) delete z;
    if(k) delete k;
    if(c) delete c;
    if(r) delete r;
    if(g) delete g;
    if(b) delete b;
    return;
}

//############################################################################
void ilda_section::tell()
{
    cout << endl;
    switch(format)
    {
        case  0: cout << "    3D frame" << endl; break;
        case  1: cout << "    2D frame" << endl; break;
        case  2: cout << "    palette"  << endl; break;
        default: cout << "    unknown"  << endl; break;
    }
    cout << "    " <<  name.c_str() << endl;
    cout << "    " << owner.c_str() << endl;

    cout << "    quantity = " <<      quantity << endl;
    cout << "    identity = " <<      identity << endl;
    cout << "    total    = " <<      total    << endl;
    cout << "    scanner  = " << (int)scanner  << endl;
    print_ilda_return(error);
    return;
}

//############################################################################
ilda_section& ilda_section::operator = (const ilda_section& section)
{
    if(section.error == ILDA_OK)
    {
        format    = section.format  ;
        quantity  = section.quantity;
        identity  = section.identity;
        total     = section.total   ;
        scanner   = section.scanner ;
        future    = 0               ;
        error     = ILDA_OK         ;

        get_mem();
        switch(format)
        {
            case  0: 
            case  1: 
                     memcpy((void*)x, (const void*)section.x, quantity * 2);
                     memcpy((void*)y, (const void*)section.y, quantity * 2);
                     memcpy((void*)z, (const void*)section.z, quantity * 2);
                     memcpy((void*)k, (const void*)section.k, quantity    );
                     memcpy((void*)c, (const void*)section.c, quantity    );
                     break;
            //----------------------------------------------------------------
            case  2: 
                     memcpy((void*)r, (const void*)section.r, quantity    );
                     memcpy((void*)g, (const void*)section.g, quantity    );
                     memcpy((void*)b, (const void*)section.b, quantity    );
                     break;
            //----------------------------------------------------------------
            default: error = ILDA_UNKNOWN_FORMAT;
                     break;
        }
    }
    return *this;
}

//############################################################################
ilda_section& ilda_section::operator += (const ilda_section& section)
{
    if(format == 2 || section.format == 2)
    {
        error = ILDA_INCOMPATIBLE_FORMATS;
        return *this;
    }

    ilda_section sum(0, "es", "LaserBoy", quantity + section.quantity + 2, 0, 0, 0);

    memcpy((void*)sum.x, (const void*)x, quantity * 2);
    memcpy((void*)sum.y, (const void*)y, quantity * 2);
    memcpy((void*)sum.z, (const void*)z, quantity * 2);
    memcpy((void*)sum.k, (const void*)k, quantity    );
    memcpy((void*)sum.c, (const void*)c, quantity    );

    sum.x[quantity] = sum.x[quantity - 1];
    sum.y[quantity] = sum.y[quantity - 1];
    sum.z[quantity] = sum.z[quantity - 1];
    sum.k[quantity] = 1;
    sum.c[quantity] = sum.c[quantity - 1];

    sum.x[quantity + 1] = section.x[0];
    sum.y[quantity + 1] = section.y[0];
    sum.z[quantity + 1] = section.z[0];
    sum.k[quantity + 1] = 1;
    sum.c[quantity + 1] = section.c[0];

    memcpy((void*)&(sum.x[quantity + 2]), (const void*)section.x, section.quantity * 2);
    memcpy((void*)&(sum.y[quantity + 2]), (const void*)section.y, section.quantity * 2);
    memcpy((void*)&(sum.z[quantity + 2]), (const void*)section.z, section.quantity * 2);
    memcpy((void*)&(sum.k[quantity + 2]), (const void*)section.k, section.quantity    );
    memcpy((void*)&(sum.c[quantity + 2]), (const void*)section.c, section.quantity    );

    *this = sum;
    return *this;
}

//############################################################################
ilda_section ilda_section::operator + (const ilda_section& section)
{
    ilda_section sum(*this);
    sum += section;
    return sum;
}

//############################################################################
char ilda_section::get_point_color(u_int &point)
{
    if(format == 2) return 0;
    if(point <  0       ) point = quantity - 1;
    if(point >= quantity) point = 0;
    return c[point];
}

//############################################################################
void ilda_section::set_point_color(u_int &point, u_char color)
{
    if(format == 2) return;
    if(point <  0       ) point = quantity - 1;
    if(point >= quantity) point = 0;
    c[point] = color;
    return;
}

//############################################################################
void ilda_section::set_points_color(u_int &point1, u_int &point2, u_char color)
{
    u_int i;
    if(format == 2) return;
    if(point1 <  0       ) point1 = quantity - 1;
    if(point1 >= quantity) point1 = 0;
    if(point2 <  0       ) point2 = quantity - 1;
    if(point2 >= quantity) point2 = 0;
    for(i = point1; i != point2; i = (i++ % quantity))
        c[i] = color;
    c[(i++ % quantity)] = color;
    return;
}

//############################################################################
void ilda_section::blank_point(u_int &point)
{
    if(format == 2) return;
    if((int)point <  0) point = quantity - 1;
    if(point >= quantity) point = 0;
    k[point] = 0x01;
    return;
}
//############################################################################
void ilda_section::blank_points(u_int &point1, u_int &point2)
{
    u_int i;
    if(format == 2) return;
    if((int)point1 <  0  ) point1 = quantity - 1;
    if(point1 >= quantity) point1 = 0;
    if((int)point2 <  0  ) point2 = quantity - 1;
    if(point2 >= quantity) point2 = 0;
    for(i = point1; i != point2; i = (i++ % quantity))
        k[i] = 0x01;
    k[(i++ % quantity)] = 0x01;
    return;
}

//############################################################################
void ilda_section::unblank_point(u_int &point)
{
    if(format == 2) return;
    if((int)point <  0  ) point = quantity - 1;
    if(point >= quantity) point = 0;
    k[point] = 0x00;
    return;
}

//############################################################################
void ilda_section::unblank_points(u_int &point1, u_int &point2)
{
    u_int i;
    if(format == 2) return;
    if((int)point1 <  0  ) point1 = quantity - 1;
    if(point1 >= quantity) point1 = 0;
    if((int)point2 <  0  ) point2 = quantity - 1;
    if(point2 >= quantity) point2 = 0;
    for(i = point1; i != point2; i = (i++ % quantity))
        k[i] = 0x00;
    k[(i++ % quantity)] = 0x00;
    return;
}

//############################################################################
void ilda_section::flip(int plane)
{
    if(format == 2) return;
    u_int i;
    switch(plane)
    {
       case 0: // X
               for(i = 0; i < quantity; i++)
                   x[i] = -x[i];
               break;                                
       // --------------------------------------------------------------------
       case 1: // Y
               for(i = 0; i < quantity; i++)
                   y[i] = -y[i];
               break;
       // --------------------------------------------------------------------
       case 3: // Z
               for(i = 0; i < quantity; i++)
                   z[i] = -z[i];
               break;
       // --------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ilda_section::quarter_turn(int plane, int turns)
{
    if(format == 2) return;
    int i, j;
    short temp;
    switch(plane)
    {
       case 0: // X Y
               for(i = 0; i < quantity; i++)
                   for(j = 0; j < turns; j++)
                   {
                       temp = -x[i];
                       x[i] = y[i];
                       y[i] = temp;
                   }
               break;                                
       // --------------------------------------------------------------------
       case 1: // Z Y
               for(i = 0; i < quantity; i++)
                   for(j = 0; j < turns; j++)
                   {
                       temp = -z[i];
                       z[i] = y[i];
                       y[i] = temp;
                   }
               break;
       // --------------------------------------------------------------------
       case 2: // X Z
               for(i = 0; i < quantity; i++)
                   for(j = 0; j < turns; j++)
                   {
                       temp = -x[i];
                       x[i] = z[i];
                       z[i] = temp;
                   }
               break;
       // --------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ilda_section::z_order_points(short int span)
{
    if(format == 2) return;
    int step = span / quantity;
    for(u_int i = 0; i < quantity; i++)
        z[i] = i * step - span / 2;
    return;
}

//############################################################################
void ilda_section::flatten_z()
{
    if(format == 2) return;
    for(u_int i = 0; i < quantity; i++)
        z[i] = 0;
    return;
}

//############################################################################
void ilda_section::concentric_sine_ripple(float amplitude, float freq, float phase)
{
    if(format == 2) return;
    for(u_int i = 0; i < quantity; i++)
        z[i] = short(amplitude * (16384.0 / pi) * sin(sqrt(float(x[i] * x[i] + y[i] * y[i])) / 32768.0 * two_pi * freq + phase));
    return;
}

//############################################################################
void ilda_section::parallel_sine_ripple(int direction, float amplitude, float freq, float phase)
{
    if(format == 2) return;
    u_int i;
    switch(direction)
    {
        case 0:
                for(i = 0; i < quantity; i++)
                    z[i] = short(amplitude * (16384.0 / pi) * sin(x[i] / 32768.0 * two_pi * freq + phase));
                break;
        case 1:
                for(i = 0; i < quantity; i++)
                    z[i] = short(amplitude * (16384.0 / pi) * sin(y[i] / 32768.0 * two_pi * freq + phase));
                break;
        case 2:
                for(i = 0; i < quantity; i++)
                    z[i] = short(amplitude * (16384.0 / pi) * sin(x[i] * y[i] / 1073741824.0 * two_pi * freq + phase));
                break;
    }
    return;
}

//############################################################################
int ilda_section::move(short  dx, short  dy, short  dz)
{
    if(format == 2) return 0;
    int  i;
    for(i = 0; i < quantity; i++)
    {
        if((x[i] + dx) > ILDA_MAX_X || (x[i] + dx) < ILDA_MIN_X)
            return 1;
        if((y[i] + dy) > ILDA_MAX_Y || (y[i] + dy) < ILDA_MIN_Y)
            return 2;
        if((z[i] + dz) > ILDA_MAX_Z || (z[i] + dz) < ILDA_MIN_Z)
            return 3;
    }
    for(i = 0; i < quantity; i++)
    {
        x[i] += dx;
        y[i] += dy;
        z[i] += dz;
    }
    return 0;
}

//############################################################################
int ilda_section::scale(float  fx, float  fy, float  fz)
{
    u_int i;
    if(format == 2) return 0;
    for(i = 0; i < quantity; i++)
    {
        if((x[i] * fx) > (float)ILDA_SHORT_MAX || (x[i] * fx) < (float)ILDA_SHORT_MIN)
            return 1;
        if((y[i] * fy) > (float)ILDA_SHORT_MAX || (y[i] * fy) < (float)ILDA_SHORT_MIN)
            return 2;
        if((z[i] * fz) > (float)ILDA_SHORT_MAX || (z[i] * fy) < (float)ILDA_SHORT_MIN)
            return 3;
    }
    for(i = 0; i < quantity; i++)
    {
        x[i] = short(x[i] * fx);
        y[i] = short(y[i] * fx);
        z[i] = short(z[i] * fx);
    }
    return 0;
}

//############################################################################
void ilda_section::rainbow_recolor(int effect, u_char c1, u_char c2, float span)
{
    if(format == 2) return;
    u_int  i, j, start, end;
    u_char color;
    switch(effect)
    {
        case 0: for(i = 0; i < quantity; i++) // span through x
                    c[i] = u_char((x[i] * span / 65536.0) * (c2 - c1)) % (c2 - c1) + c1;
                break;
       //---------------------------------------------------------------------
        case 1: for(i = 0; i < quantity; i++) // span through y
                    c[i] = u_char((y[i] * span / 65536.0) * (c2 - c1)) % (c2 - c1) + c1;
                break;
       //---------------------------------------------------------------------
        case 2: for(i = 0; i < quantity; i++) // span through z
                    c[i] = u_char((z[i] * span / 65536.0) * (c2 - c1)) % (c2 - c1) + c1;
                break;
       //---------------------------------------------------------------------
        case 3: for(i = 0; i < quantity; i++) // span through points
                    c[i] = u_char((i * span / float(quantity)) * (c2 - c1)) % (c2 - c1) + c1;
                break;
       //---------------------------------------------------------------------
        case 4: for(i = 0; i < quantity; i++) // index palette through points
                    c[i] = i % (c2 - c1) + c1;
                break;
       //---------------------------------------------------------------------
        case 5: for(i = 1; i <= number_of_shapes(); i++) // index through shapes
                {
                    color = ((i - 1) % (c2 - c1)) + c1;
                    find_shape(i, start, end);
                    for(j = start; j <= end; j++)
                        c[j] = color;
                }
                break;
       //---------------------------------------------------------------------
        case 6: for(i = 1; i <= number_of_shapes(); i++) // span through shapes
                {
                    color = u_char(((i - 1) * span / number_of_shapes()) * (c2 - c1)) % (c2 - c1) + c1;
                    find_shape(i, start, end);
                    for(j = start; j <= end; j++)
                        c[j] = color;
                }
                break;
       //---------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ilda_section::palette_cycle(u_char c1, u_char c2, int step)
{
    if(format == 2) return;
    for(u_int i = 0; i < quantity; i++)
        c[i] = ((c[i] - c1 + step) % (c2 - c1)) + c1;
    return;
}

//############################################################################
void ilda_section::rotate_point(short   x, short   y, short   z,
                                float& rx, float& ry, float& rz,
                                float  ax, float  ay, float  az)
{
    float sin_of_ax = sin(ax),
          sin_of_ay = sin(ay),
          sin_of_az = sin(az),
          cos_of_ax = cos(ax),
          cos_of_ay = cos(ay),
          cos_of_az = cos(az);
                    
    ry =  y * cos_of_ax -  z * sin_of_ax;
    rz =  z * cos_of_ax +  y * sin_of_ax; // rotate around x
    rz = rz * cos_of_ay -  x * sin_of_ay;
    rx =  x * cos_of_ay + (z * cos_of_ax +  y * sin_of_ax) * sin_of_ay; // rotate around y
    rx = rx * cos_of_az - (y * cos_of_ax -  z * sin_of_ax) * sin_of_az;
    ry = ry * cos_of_az + (x * cos_of_ay + (z * cos_of_ax +  y * sin_of_ax) * sin_of_ay) * sin_of_az; // rotate around z
    return;
}

//############################################################################
void ilda_section::rotate_point_around_point(short   x, short   y, short   z,
                                             short  cx, short  cy, short  cz,
                                             float& rx, float& ry, float& rz,
                                             float  ax, float  ay, float  az)
{
    float sin_of_ax = sin(ax),
          sin_of_ay = sin(ay),
          sin_of_az = sin(az),
          cos_of_ax = cos(ax),
          cos_of_ay = cos(ay),
          cos_of_az = cos(az);
    x -= cx;
    y -= cy;
    z -= cz;    
    ry =  y * cos_of_ax -  z * sin_of_ax;
    rz =  z * cos_of_ax +  y * sin_of_ax; // rotate around x
    rz = rz * cos_of_ay -  x * sin_of_ay;
    rx =  x * cos_of_ay + (z * cos_of_ax +  y * sin_of_ax) * sin_of_ay; // rotate around y
    rx = rx * cos_of_az - (y * cos_of_ax -  z * sin_of_ax) * sin_of_az;
    ry = ry * cos_of_az + (x * cos_of_ay + (z * cos_of_ax +  y * sin_of_ax) * sin_of_ay) * sin_of_az; // rotate around z
    rx += cx;
    ry += cy;
    rz += cz;
    return;
}

//############################################################################
void ilda_section::scale_point_around_point(short   x, short   y, short   z,
                                            short  cx, short  cy, short  cz,
                                            float& sx, float& sy, float& sz,
                                            float  mx, float  my, float  mz)
{
    x -= cx;
    y -= cy;
    z -= cz;    
    sx = x * mx;
    sy = y * my;
    sz = z * mz;
    sx += cx;
    sy += cy;
    sz += cz;
}

//############################################################################
void ilda_section::rotate_all_around_origin(float  ax, float  ay, float  az)
{
    if(format == 2) return;
    float rx, ry, rz;
    for(u_int i = 0; i < quantity; i++)
    {
        rotate_point(x[i], y[i], z[i], rx, ry, rz, ax, ay, az);
        x[i] = short(rx);
        y[i] = short(ry);
        z[i] = short(rz);
    }
    return;
}

//############################################################################
u_int ilda_section::number_of_shapes()
{
    if(format == 2) return 0;
    u_int i,
          in_shape = 0,
          shapes   = 0;

    for(i = 0; i < quantity; i++)
    {
          if(!k[i] && !in_shape)
          {
              shapes++;
              in_shape = 1;
          }
          else if(k[i])
              in_shape = 0;
    }
    return shapes;
}
    
//############################################################################
void ilda_section::find_shape(u_int shape, u_int& start, u_int& end)
{
    if(format == 2) return;
    u_int i,
          shape_num = 0;

    shape %= number_of_shapes();

    for(i = 0; i < quantity; i++)
    {
          if(!k[i])
          {
              start = (i) ? (i - 1) : (0);
              while(!k[i] && i < quantity)
                  i++;
              if(i < quantity)
                  end = i;
              else
                  end = quantity - 1;
              if(shape == shape_num)
                  return;
              shape_num++;
          }
    }
    return;
}

//############################################################################
void ilda_section::find_shape_center(u_int shape, short& cx, short& cy, short& cz)
{
    u_int i,
          start,
          end;
    int   max_x = -40000, // more negative than any short
          min_x =  40000, // more positive than any short
          max_y = -40000,
          min_y =  40000,
          max_z = -40000,
          min_z =  40000;
    find_shape(shape, start, end);
    for(i = start; i <= end; i++)
    {
        if(x[i] >= max_x) max_x = x[i];
        if(x[i] <= min_x) min_x = x[i];
        if(y[i] >= max_y) max_y = y[i];
        if(y[i] <= min_y) min_y = y[i];
        if(z[i] >= max_z) max_z = z[i];
        if(z[i] <= min_z) min_z = z[i];
    }
    cx = short(max_x - ((max_x - min_x) / 2));
    cy = short(max_y - ((max_y - min_y) / 2));
    cz = short(max_z - ((max_z - min_z) / 2));
    return;
}

//############################################################################
int ilda_section::move_shape(u_int shape, short dx, short dy, short dz)
{
    if(format == 2) return 0;
    u_int  i,
           start,
           end;
    find_shape(shape, start, end);
    for(i = start; i <= end; i++)
    {
        if((x[i] + dx) > ILDA_MAX_X || (x[i] + dx) < ILDA_MIN_X)
            return 1;
        if((y[i] + dy) > ILDA_MAX_Y || (y[i] + dy) < ILDA_MIN_Y)
            return 2;
        if((z[i] + dz) > ILDA_MAX_Z || (z[i] + dz) < ILDA_MIN_Z)
            return 3;
    }
    for(i = start; i <= end; i++)
    {
        x[i] += dx;
        y[i] += dy;
        z[i] += dz;
    }
    return 0;
}

//############################################################################
int ilda_section::rotate_shape(u_int shape, float ax, float ay, float az)
{
    if(format == 2) return 0;
    u_int  i,
           start,
           end;
    short  cx = 0,
           cy = 0,
           cz = 0;
    float  rx, ry, rz;

    find_shape(shape, start, end);
    find_shape_center(shape, cx, cy, cz);

    for(i = start; i <= end; i++)
    {
        rotate_point_around_point(x[i], y[i], z[i], cx, cy, cz, rx, ry, rz, ax, ay, az);
        if(rx > ILDA_MAX_X || rx < ILDA_MIN_X)
            return 1;
        if(ry > ILDA_MAX_Y || ry < ILDA_MIN_Y)
            return 2;
        if(rz > ILDA_MAX_Z || rz < ILDA_MIN_Z)
            return 3;
    }
    for(i = start; i <= end; i++)
    {
        rotate_point_around_point(x[i], y[i], z[i], cx, cy, cz, rx, ry, rz, ax, ay, az);
        x[i] = short(rx);
        y[i] = short(ry);
        z[i] = short(rz);
    }
    return 0;
}

//############################################################################
int ilda_section::rotate_shape_around_origin(u_int shape, float ax, float ay, float az)
{
    if(format == 2) return 0;
    u_int  i,
           start,
           end;
    float  rx, ry, rz;
    find_shape(shape, start, end);
    for(i = start; i <= end; i++)
    {
        rotate_point(x[i], y[i], z[i], rx, ry, rz, ax, ay, az);
        if(rx > ILDA_MAX_X || rx < ILDA_MIN_X)
            return 1;
        if(ry > ILDA_MAX_Y || ry < ILDA_MIN_Y)
            return 2;
        if(rz > ILDA_MAX_Z || rz < ILDA_MIN_Z)
            return 3;
    }
    for(i = start; i <= end; i++)
    {
        rotate_point(x[i], y[i], z[i], rx, ry, rz, ax, ay, az);
        x[i] = short(rx);
        y[i] = short(ry);
        z[i] = short(rz);
    }
    return 0;
}

//############################################################################
int ilda_section::scale_shape(u_int shape, float mx, float my, float mz)
{
    if(format == 2) return 0;
    u_int  i,
           start,
           end;
    short  cx = 0,
           cy = 0,
           cz = 0;
    float  sx, sy, sz;

    find_shape        (shape, start, end);
    find_shape_center (shape, cx, cy, cz);

    for(i = start; i <= end; i++)
    {
        scale_point_around_point(x[i], y[i], z[i], cx, cy, cz, sx, sy, sz, mx, my, mz);
        if(sx > ILDA_MAX_X || sx < ILDA_MIN_X)
            return 1;
        if(sy > ILDA_MAX_Y || sy < ILDA_MIN_Y)
            return 2;
        if(sz > ILDA_MAX_Z || sz < ILDA_MIN_Z)
            return 3;
    }
    for(i = start; i <= end; i++)
    {
        scale_point_around_point(x[i], y[i], z[i], cx, cy, cz, sx, sy, sz, mx, my, mz);
        x[i] = short(sx);
        y[i] = short(sy);
        z[i] = short(sz);
    }
    return 0;
}

//############################################################################
int ilda_section::scale_shape_around_origin(u_int shape, float mx, float my, float mz)
{
    if(format == 2) return 0;
    u_int  i,
           start,
           end;
    float  sx, sy, sz;
    find_shape(shape, start, end);
    for(i = start; i <= end; i++)
    {
        sx = x[i] * mx;
        if(sx > ILDA_MAX_X || sx < ILDA_MIN_X)
            return 1;
        sy = y[i] * my;
        if(sy > ILDA_MAX_Y || sy < ILDA_MIN_Y)
            return 2;
        sz = z[i] * mz;
        if(sz > ILDA_MAX_Z || sz < ILDA_MIN_Z)
            return 3;
    }
    for(i = start; i <= end; i++)
    {
        x[i] = short(x[i] * mx);
        y[i] = short(y[i] * my);
        z[i] = short(z[i] * mz);
    }
    return 0;
}

//############################################################################
void ilda_section::wiggle_shapes(int plane, int max)
{
    if(format == 2) return;
    u_int  shape;
    int    dx, dy, dz;

    switch(plane)
    {
        case 0: // X Y
                for(shape = 0; shape < number_of_shapes(); shape++)
                {                  
                    do
                    {
                        dx = int((ezfb_random_01() - 0.5) * 2.0 * max);
                        dy = int((ezfb_random_01() - 0.5) * 2.0 * max);                      
                    } while(move_shape(shape, dx, dy, 0));
                }
                break;
        //--------------------------------------------------------------------
        case 1: // Z Y
                for(shape = 0; shape < number_of_shapes(); shape++)
                {                  
                    do
                    {
                        dz = int((ezfb_random_01() - 0.5) * 2.0 * max);
                        dy = int((ezfb_random_01() - 0.5) * 2.0 * max);                      
                    } while(move_shape(shape, 0, dy, dz));
                }
                break;
        //--------------------------------------------------------------------
        case 2: // X Z
                for(shape = 0; shape < number_of_shapes(); shape++)
                {                  
                    do
                    {
                        dx = int((ezfb_random_01() - 0.5) * 2.0 * max);
                        dz = int((ezfb_random_01() - 0.5) * 2.0 * max);                      
                    } while(move_shape(shape, dx, 0, dz));
                }
                break;
        //--------------------------------------------------------------------
        case 3: // X Y Z
                for(shape = 0; shape < number_of_shapes(); shape++)
                {                  
                    do
                    {
                        dx = int((ezfb_random_01() - 0.5) * 2.0 * max);
                        dy = int((ezfb_random_01() - 0.5) * 2.0 * max);
                        dz = int((ezfb_random_01() - 0.5) * 2.0 * max);                      
                    } while(move_shape(shape, dx, dy, dz));
                }
                break;
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ilda_section::roto_wiggle_shapes(int plane, float max)
{
    if(format == 2) return;
    u_int shape;
    float ax, ay, az;
    switch(plane)
    {
      case 0: // X Y
                for(shape = 0; shape < number_of_shapes(); shape++)
                {                  
                    do
                    {
                        az = (ezfb_random_01() - 0.5) * 2.0 * max;
                    } while(rotate_shape(shape, 0, 0, az));
                }
                break;
      //----------------------------------------------------------------------
      case 1: // Z Y
                for(shape = 0; shape < number_of_shapes(); shape++)
                {                  
                    do
                    {
                        ax = (ezfb_random_01() - 0.5) * 2.0 * max;
                    } while(rotate_shape(shape, ax, 0, 0));
                }
                break;
      //----------------------------------------------------------------------
      case 2: // X Z
                for(shape = 0; shape < number_of_shapes(); shape++)
                {                  
                    do
                    {
                        ay = (ezfb_random_01() - 0.5) * 2.0 * max;
                    } while(rotate_shape(shape, 0, ay, 0));
                }
                break;
      //----------------------------------------------------------------------
      case 3: // X Y Z
                for(shape = 0; shape < number_of_shapes(); shape++)
                {                  
                    do
                    {
                        ax = (ezfb_random_01() - 0.5) * 2.0 * max;
                        ay = (ezfb_random_01() - 0.5) * 2.0 * max;
                        az = (ezfb_random_01() - 0.5) * 2.0 * max;
                    } while(rotate_shape(shape, ax, ay, az));
                }
                break;
      //----------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ilda_section::falling_shapes(int plane, int max)
{
    if(format == 2) return;
    u_int i,
          shape,
          start,
          end;
    int   //dx,
          dy,
          //dz,
          //out_x = 0,
          //out_z = 0,
          out_y = 0;

    for(shape = 0; shape < number_of_shapes(); shape++)
        switch(plane)
        {
            case 0:
                     //dx = int((ezfb_random_01() - 0.5) * 2.0 * max);
                    dy = int((ezfb_random_01() + 1.0) * 2.0 * max); // all negative in Y
                    //dz = int((ezfb_random_01() - 0.5) * 2.0 * max);
                    find_shape(shape, start, end);
                    for(i = start; i <= end; i++)
                    {
                        //if((x[i] + dx) > ILDA_MAX_X || (x[i] + dx) < ILDA_MIN_X) out_x = 1;
                        if((y[i] + dy) > ILDA_MAX_Y || (y[i] + dy) < ILDA_MIN_Y) out_y = 1;
                        //if((z[i] + dz) > ILDA_MAX_Z || (z[i] + dz) < ILDA_MIN_Z) out_z = 1;
                    }
                    //if(out_x)
                    //    for(i = start; i <= end; i++)
                    //        x[i] = -x[i];
                    if(out_y)
                        for(i = start; i <= end; i++)
                            y[i] = -y[i];
                    //if(out_z)
                    //    for(i = start; i <= end; i++)
                    //        z[i] = -z[i];
                    for(i = start; i <= end; i++)
                    {
                        //if(!out_x) x[i] += dx;
                        if(!out_y) y[i] += dy;
                        //if(!out_z) z[i] += dz;
                    }
                    //out_x = 0;
                    out_y = 0;
                    //out_z = 0;
                    break;
        //--------------------------------------------------------------------
        }
    //roto_wiggle_shapes(3, pi/50.0);
    return;
}

//############################################################################
void ilda_section::from_ifstream(ifstream& in)
{
    get_header (in);
    get_data   (in);
    return         ;
}    
    
//############################################################################
void ilda_section::get_header(ifstream& in)
{
    int i;
    char p, q;

    name.erase();
    owner.erase();

    if(!in.eof())
    {
        in.get(p); if(p != 'I') { error = ILDA_BAD_HEADER; return; }
        in.get(p); if(p != 'L') { error = ILDA_BAD_HEADER; return; }
        in.get(p); if(p != 'D') { error = ILDA_BAD_HEADER; return; }
        in.get(p); if(p != 'A') { error = ILDA_BAD_HEADER; return; }
        in.get(p); if(p !=  0 ) { error = ILDA_BAD_HEADER; return; }
        in.get(p); if(p !=  0 ) { error = ILDA_BAD_HEADER; return; }
        in.get(p); if(p !=  0 ) { error = ILDA_BAD_HEADER; return; }

        in.get(format);

        for(i = 0; i < 8; i++)
        {
            in.get(p);
            name += p;
        }
        for(i = 0; i < 8; i++)
        {
            in.get(p);
            owner += p;
        }

        in.get(p); in.get(q);
        quantity = ((p & 0x00ff) << 8 | (q & 0x00ff));

        in.get(p); in.get(q);
        identity = ((p & 0x00ff) << 8 | (q & 0x00ff));

        in.get(p); in.get(q);
        total    = ((p & 0x00ff) << 8 | (q & 0x00ff));

        in.get(scanner);
        in.get(future);

        if(in.bad()) { error = ILDA_BAD_HEADER; return; }
    }
    error = ILDA_OK;
    return;
}
            
//############################################################################
void ilda_section::get_data(ifstream& in)
{
    if(!in.eof() && (quantity != 0))
    {
        u_int i;
        char  p, q;
        delete_mem();
        switch(format)
        {
            case  0:
                     x = new short[quantity];
                     y = new short[quantity];
                     z = new short[quantity];
                     k = new char [quantity];
                     c = new char [quantity];
                     for(i = 0; i < quantity; i++)
                     {   
                         in.get(p); in.get(q);
                         x[i] = ((p & 0x00ff) << 8 | (q & 0x00ff));

                         in.get(p); in.get(q);
                         y[i] = ((p & 0x00ff) << 8 | (q & 0x00ff));

                         in.get(p); in.get(q);
                         z[i] = ((p & 0x00ff) << 8 | (q & 0x00ff));

                         in.get(k[i]); in.get(c[i]); k[i] = (k[i] & 64);
                     }         
                     break;
            //----------------------------------------------------------------
            case  1:
                     x = new short[quantity];
                     y = new short[quantity];
                     k = new char [quantity];
                     c = new char [quantity];
                     for(i = 0; i < quantity; i++)
                     {   
                         in.get(p); in.get(q);
                         x[i] = ((p & 0x00ff) << 8 | (q & 0x00ff));

                         in.get(p); in.get(q);
                         y[i] = ((p & 0x00ff) << 8 | (q & 0x00ff));

                         in.get(k[i]); in.get(c[i]); k[i] = (k[i] & 64);
                     }         
                     break;
            //----------------------------------------------------------------
            case  2: r = new char [quantity];
                     g = new char [quantity];
                     b = new char [quantity];
                     for(i = 0; i < quantity; i++)
                     {   
                         in.get(r[i]);
                         in.get(g[i]);
                         in.get(b[i]);
                     }         
                     break;
            //----------------------------------------------------------------
        }
        if(in.bad()) { error = ILDA_EOF_ERROR; return; }
    }

    error = ILDA_OK;
    return;
}

//############################################################################
void ilda_section::to_ofstream(ofstream& out)
{
    write_header(out);
    write_data(out);
    return;
}

//############################################################################
void ilda_section::write_header(ofstream& out)
{
    int i;
    out.put('I');
    out.put('L');
    out.put('D');
    out.put('A');
    out.put( 0 );
    out.put( 0 );
    out.put( 0 );
    out.put(format);
    for(i = 0; i < 8; i++)
        out.put(name[i]);
    for(i = 0; i < 8; i++)
        out.put(owner[i]);

    out.put((char)((quantity & 0xff00) >> 8));
    out.put((char) (quantity & 0x00ff)      );

    out.put((char)((identity & 0xff00) >> 8));
    out.put((char) (identity & 0x00ff)      );

    out.put((char)((total    & 0xff00) >> 8));
    out.put((char) (total    & 0x00ff)      );

    out.put(scanner);
    out.put( 0 ); // future
    return;
}

//############################################################################
void ilda_section::write_data(ofstream& out)
{
    u_int i;
    switch(format)
    {
        case  0: for(i = 0; i < quantity; i++)
                 {   
                     out.put((char)((x[i] & 0xff00) >> 8));
                     out.put((char) (x[i] & 0x00ff)      );
                     out.put((char)((y[i] & 0xff00) >> 8));
                     out.put((char) (y[i] & 0x00ff)      );
                     out.put((char)((z[i] & 0xff00) >> 8));
                     out.put((char) (z[i] & 0x00ff)      );
                     out.put((k[i]) ? (64) : (0));
                     out.put(c[i]);                     
                 }         
                 break;
            //----------------------------------------------------------------
        case  1: for(i = 0; i < quantity; i++)
                 {   
                     out.put((char)((x[i] & 0xff00) >> 8));
                     out.put((char) (x[i] & 0x00ff)      );
                     out.put((char)((y[i] & 0xff00) >> 8));
                     out.put((char) (y[i] & 0x00ff)      );
                     out.put((k[i]) ? (64) : (0));
                     out.put(c[i]);                     
                 }         
                 break;
            //----------------------------------------------------------------
        case  2: for(i = 0; i < quantity; i++)
                 {   
                     out.put(r[i]);
                     out.put(g[i]);
                     out.put(b[i]);
                 }         
                 break;
            //----------------------------------------------------------------
    }
    return;
}

//############################################################################






//############################################################################
#ifdef __WITH_BMPTOOLS__
//############################################################################
void  ilda_section::render_as_bmp(int size, float ax, float ay, float az)
{
    if(quantity && ((format == 0) || (format == 1)))
    {
        char file_name[14];
        int   i;
        float rx1, ry1, rz1,
              rx2, ry2, rz2,
              scale = (bmp->yres / 65536.0);
        sprintf(file_name, "%08d.bmp", identity);
        BMP_8 bmp(size, size, file_name);
        bmp.palette();
        bmp.report();
        for(i = 1; i < quantity; i++)
        {
            if(!k[i])
            {
                if(ax || ay || az)
                {
                    rotate_point(x[i - 1], y[i - 1], z[i - 1], rx1, ry1, rz1, ax, ay, az);
                    rotate_point(x[i    ], y[i    ], z[i    ], rx2, ry2, rz2, ax, ay, az);
                }
                else
                {
                    rx1 = x[i - 1];
                    ry1 = y[i - 1];
                    rz1 = z[i - 1];
                    rx2 = x[i    ];
                    ry2 = y[i    ];
                    rz2 = z[i    ];
                }
                bmp.line( (int)(rx1 * scale) + bmp.x / 2,
                          (int)(ry1 * scale) + bmp.x / 2,
                          (int)(rx2 * scale) + bmp.x / 2,
                          (int)(ry2 * scale) + bmp.x / 2,
                          c[i]                            );

            }
        }
        bmp.save();
    }
    return;
}
//############################################################################
#endif
//############################################################################












//############################################################################
#ifdef __WITH_EZFB__
//############################################################################
ilda_section::ilda_section(ezfb_bitmap* bmp, short id)
                          :format    (2      ),
                           quantity  (256    ),
                           identity  (id     ),
                           total     (0      ),
                           scanner   (0      ),
                           future    (0      ),
                           x         (NULL   ),
                           y         (NULL   ),
                           z         (NULL   ),
                           k         (NULL   ),
                           c         (NULL   ),
                           r         (NULL   ),
                           g         (NULL   ),
                           b         (NULL   ),
                           error     (ILDA_OK)
{
    int i;
    get_mem();
    for(i = 0; i < 256; i++)
    {
        r[i] = ((bmp->palette[i] & 0x0000ff00) >> 8 );
        g[i] = ((bmp->palette[i] & 0x00ff0000) >> 16);
        b[i] = ((bmp->palette[i] & 0xff000000) >> 24);
    }
}

//############################################################################
void ilda_section::render_in_ezfb(struct ezfb* fb, float ax, float ay, float az)
{
    if(quantity && ((format == 0) || (format == 1)))
    {
        int   i;
        float rx1, ry1, rz1,
              rx2, ry2, rz2,
              scale = (fb->Var.yres / 65536.0);
        for(i = 1; i < quantity; i++)
        {
            if(!k[i])
            {
                if(ax || ay || az)
                {
                    rotate_point(x[i - 1], y[i - 1], z[i - 1], rx1, ry1, rz1, ax, ay, az);
                    rotate_point(x[i    ], y[i    ], z[i    ], rx2, ry2, rz2, ax, ay, az);
                }
                else
                {
                    rx1 = x[i - 1];
                    ry1 = y[i - 1];
                    rz1 = z[i - 1];
                    rx2 = x[i    ];
                    ry2 = y[i    ];
                    rz2 = z[i    ];
                }
                ezfb_put_line(fb,                 (int)(x[i - 1] * scale) + fb->mid_y
                          , fb->Var.yres - ((int)(y[i - 1] * scale) + fb->mid_y)
                          ,                 (int)(x[i    ] * scale) + fb->mid_y
                          , fb->Var.yres - ((int)(y[i    ] * scale) + fb->mid_y)
                          , c[i]
                       );
            }
        }
    }
    else if(format == 2)
    {
        u_int i, j, k = 0, x, y;
        for(i = 0; i < 16; i++)
            for(j = 0; j < 16; j++)
            {
                for(y = 0; y < (fb->Var.yres / 16 - 1); y++)
                    for(x = 0; x < (fb->Var.yres / 16 - 1); x++)
                        ezfb_put_pixel(fb, j * (fb->Var.xres / 16) + x, i * (fb->Var.yres / 24) + y, k);
                k++;
            }
    }
}

//############################################################################
void ilda_section::render_in_ezfb_bitmap(struct ezfb_bitmap* bmp, float ax, float ay, float az)
{
    bmp_clear(bmp, bmp->black);
    if(quantity && ((format == 0) || (format == 1)))
    {
        int   i;
        float rx1, ry1, rz1,
              rx2, ry2, rz2,
              scale = (bmp->yres / 65536.0);
        bmp_put_rectangle(bmp, 0, 0, bmp->yres - 1, bmp->yres - 1, bmp->white, 0);
        for(i = 1; i < quantity; i++)
        {
            if(!k[i])
            {
                if(ax || ay || az)
                {
                    rotate_point(x[i - 1], y[i - 1], z[i - 1], rx1, ry1, rz1, ax, ay, az);
                    rotate_point(x[i    ], y[i    ], z[i    ], rx2, ry2, rz2, ax, ay, az);
                }
                else
                {
                    rx1 = x[i - 1];
                    ry1 = y[i - 1];
                    rz1 = z[i - 1];
                    rx2 = x[i    ];
                    ry2 = y[i    ];
                    rz2 = z[i    ];
                }
                bmp_put_line(bmp,              (int)(rx1 * scale) + bmp->yres / 2
                            , bmp->yres - ((int)(ry1 * scale) + bmp->yres / 2)
                            ,              (int)(rx2 * scale) + bmp->yres / 2
                            , bmp->yres - ((int)(ry2 * scale) + bmp->yres / 2)
                            , c[i]
                        );
            }
        }
    }
    else if(format == 2)
    {
        u_int i, j, k = 0, x, y;
        for(i = 0; i < 16; i++)
            for(j = 0; j < 16; j++)
            {
                for(y = 0; y < (bmp->yres / 16 - 1); y++)
                    for(x = 0; x < (bmp->yres / 16 - 1); x++)
                        bmp_put_pixel(bmp, j * (bmp->yres / 16) + x, i * (bmp->yres / 16) + y, k);
                k++;
                cout << '.';
            }
    }
    int factor = (bmp->yres / 384);
    bmp_printf(bmp, bmp->yres + 10, 256 * factor + 8 + 45 * factor, bmp->white, bmp->black, 0, factor, "%d", identity);
    return;
}
    
//############################################################################
void ilda_section::render_in_ezfb_bitmap_point_select(  ezfb_bitmap* bmp,
                                                        u_int&       point1,
                                                        u_int&       point2,
                                                        float        ax,
                                                        float        ay,
                                                        float        az
                                                     )
{
    if(quantity && ((format == 0) || (format == 1)))
    {
        float rx1, ry1, rz1, rx2, ry2, rz2,
              scale = (bmp->yres / 65536.0);
        
        if(point1 <  0       ) point1 = quantity - 1;
        if(point1 >= quantity) point1 = 0;
        if(point2 <  0       ) point2 = quantity - 1;
        if(point2 >= quantity) point2 = 0;
                
        render_in_ezfb_bitmap(bmp, ax, ay, az);

        if(ax || ay || az)
        {
            rotate_point(x[point1], y[point1], z[point1], rx1, ry1, rz1, ax, ay, az);
            rotate_point(x[point2], y[point2], z[point2], rx2, ry2, rz2, ax, ay, az);
        }
        else
        {
            rx1 = x[point1];
            ry1 = y[point1];
            rz1 = z[point1];
            rx2 = x[point2];
            ry2 = y[point2];
            rz2 = z[point2];
        }
        //--------------------------------------------------------------------
        // start point select cursor
        bmp_put_rectangle(bmp,         (int)((rx1 - 1000) * scale) + bmp->yres / 2,
                      bmp->yres - ((int)((ry1 - 1000) * scale) + bmp->yres / 2),
                                   (int)((rx1 + 1000) * scale) + bmp->yres / 2,
                      bmp->yres - ((int)((ry1 + 1000) * scale) + bmp->yres / 2),
                      (k[point1]) ? (bmp->white) : (c[point1]), 0
                     );
        //--------------------------------------------------------------------
        // end point select cursor
        bmp_put_line     (bmp,         (int)((rx2       ) * scale) + bmp->yres / 2,
                      bmp->yres - ((int)((ry2 - 1000) * scale) + bmp->yres / 2),
                                   (int)((rx2       ) * scale) + bmp->yres / 2,
                      bmp->yres - ((int)((ry2 + 1000) * scale) + bmp->yres / 2),
                      (k[point2]) ? (bmp->white) : (c[point2])
                     );
        bmp_put_line     (bmp,         (int)((rx2 + 1000) * scale) + bmp->yres / 2,
                      bmp->yres - ((int)((ry2       ) * scale) + bmp->yres / 2),
                                   (int)((rx2 - 1000) * scale) + bmp->yres / 2,
                      bmp->yres - ((int)((ry2       ) * scale) + bmp->yres / 2),
                      (k[point2]) ? (bmp->white) : (c[point2])
                     );
        bmp_put_line     (bmp,         (int)((rx2 - 1000) * scale) + bmp->yres / 2,
                      bmp->yres - ((int)((ry2 - 1000) * scale) + bmp->yres / 2),
                                   (int)((rx2 + 1000) * scale) + bmp->yres / 2,
                      bmp->yres - ((int)((ry2 + 1000) * scale) + bmp->yres / 2),
                      (k[point2]) ? (bmp->white) : (c[point2])
                     );
        bmp_put_line     (bmp,         (int)((rx2 + 1000) * scale) + bmp->yres / 2,
                      bmp->yres - ((int)((ry2 - 1000) * scale) + bmp->yres / 2),
                                   (int)((rx2 - 1000) * scale) + bmp->yres / 2,
                      bmp->yres - ((int)((ry2 + 1000) * scale) + bmp->yres / 2),
                      (k[point2]) ? (bmp->white) : (c[point2])
                     );
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ilda_section::render_palette_in_ezfb_bitmap(ezfb_bitmap* bmp, u_char selected_index)
{
    int   i, j;
    int   factor = (bmp->yres / 384);

    for(i = 0; i < 256; i++)
        for(j = 0; j < factor; j++)
            bmp_put_line(bmp, bmp->yres, i * factor + j, bmp->yres + 50, i * factor + j, i);

    bmp_put_rectangle(bmp, bmp->yres + 70, 0, bmp->yres + 120, 256 * factor, selected_index, 1);

    for(j = 0; j < factor; j++)
        bmp_put_line(bmp, bmp->yres + 40, selected_index * factor + j, bmp->yres + 80, selected_index * factor + j, bmp->white);

    bmp_printf(bmp, bmp->yres + 10, 256 * factor + 8              , bmp->white, bmp->black, 0, factor, "i : %d", selected_index);
    bmp_printf(bmp, bmp->yres + 10, 256 * factor + 8 +  9 * factor, bmp->white, bmp->black, 0, factor, "r : %d", ((bmp->palette[selected_index] & 0x0000ff00) >>  8));
    bmp_printf(bmp, bmp->yres + 10, 256 * factor + 8 + 18 * factor, bmp->white, bmp->black, 0, factor, "g : %d", ((bmp->palette[selected_index] & 0x00ff0000) >> 16));
    bmp_printf(bmp, bmp->yres + 10, 256 * factor + 8 + 27 * factor, bmp->white, bmp->black, 0, factor, "b : %d", ((bmp->palette[selected_index] & 0xff000000) >> 24));
}

//############################################################################
#endif



//############################################################################
void ilda_section::render_raw(ofstream& out)
{
    if(quantity && ((format == 0) || (format == 1)))
    {
        int i, j, k = (int)ceil((float)9600.0 / quantity);

        for(j = 0; j < k; j++)
            for(i = 0; i < quantity; i++)
            {
                out.put( x[i] & 0x00ff      );
                out.put((x[i] & 0xff00) >> 8);
                out.put( y[i] & 0x00ff      );
                out.put((y[i] & 0xff00) >> 8);
            }
    }
}

//############################################################################
// class ilda
//############################################################################
ilda& ilda::operator += (const ilda& ild)
{
    for(u_int i = 0; i < ild.num_sections; i++)
        append_section(ild.sections[i]);
    return *this;
}

//############################################################################
ilda ilda::operator + (const ilda& ild)
{
    ilda sum(*this);
    for(u_int i = 0; i < ild.num_sections; i++)
        sum.append_section(ild.sections[i]);
    return sum;
}

//############################################################################
ilda_section& ilda::operator [] (u_int index)
{
    if(index >= 0 && index < num_sections)
        return sections[index];
    
    cout << "\nERROR! index " << index << " is out of bounds";
    return dud;
}



//############################################################################
ilda_section& ilda::last_section()
{
    return sections[number_of_sections() - 1];
}

//############################################################################
int ilda::from_file(string file)
{
    int ret = ILDA_OK;
    ifstream in(file.c_str(), ios::in | ios::binary);
    if(in.is_open())
    {
        file_name = file;
        ret = from_ifstream(in);
        in.close();
    }
    else
    {
        cout << endl << file << " failed to open for input" << endl;
        ret = ILDA_FILE_OPEN_FAILED;
    }
    return ret;
}

//############################################################################
int ilda::from_ifstream(ifstream& in)
{
    int ret = (int)ILDA_OK;
    int some_data_accepted = 0;
    ilda_section section;
    sections.clear();
    num_sections  = 0;
    num_frames    = 0;
    num_2D_frames = 0;
    num_3D_frames = 0;
    num_palettes  = 0;
    while(!in.eof() && ret == ILDA_OK)
    {
        section.from_ifstream(in);
        if(section.quantity && (section.error == ILDA_OK))
        {
            append_section(section);
            some_data_accepted = 1;
        }
        ret |= section.error;
    }
    if(ret && some_data_accepted)
    {
        print_ilda_return(ret);
        cout << endl << "an error occurred" << endl;
        ret = 0;
    }
    return ret;
} 

//############################################################################
void ilda::append_section(ilda_section section)
{
    sections.push_back(section);
    num_sections++;
    switch(section.format)
    {
        case  0:
                 num_frames++;
                 num_3D_frames++;
                 section.identity = num_frames;
                 break;

        case  1:
                 num_frames++;
                 num_2D_frames++;
                 section.identity = num_frames;
                 break;

        case  2: 
                 num_palettes++;
                 break;
    }
    return;
}

//############################################################################
void ilda::append_section()
{
    append_section(dud);
    return;
}

//############################################################################
void ilda::renumber_frames()
{
    u_int i, j = 0;
    char name[7];
    for(i = 0; i < num_sections; i++)
    {
        switch(sections[i].format)
        {
            case  0:
            case  1:
                     sections[i].identity = j++;
                     sections[i].total = num_frames;
                     break;

            case  2: 
                     break;
        }
        sprintf(name, "%06d", sections[i].identity);
        sections[i].name  = "ES" + string(name);
        sections[i].owner = "LaserBoy";
    }
    return;
}

//############################################################################
void ilda::reverse()
{
    vector<ilda_section> reverse;
    for(u_int i = number_of_sections() - 1; ((int)i >= 0) ; i--)
        reverse.push_back(sections[i]);
    sections = reverse;
    renumber_frames();
    return;
}

//############################################################################
void ilda::save_as(string file)
{
    ofstream out(file.c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        to_ofstream(out);
        out.close();
    }
    else
        cout << endl << file << " failed to open for output" << endl;
}

//############################################################################
void ilda::to_ofstream(ofstream& out)
{
    renumber_frames();
    for(u_int i = 0; i < num_sections; i++)
        if(sections[i].quantity)
            sections[i].to_ofstream(out);
    return;
}

//############################################################################
void ilda::recount()
{
    num_sections = number_of_sections();
    for(u_int i = 0; i < number_of_sections() ; i++)
    {
        switch(sections[i].format)
        {
            case  0:
                     num_frames++;
                     num_3D_frames++;
                     break;

            case  1:
                     num_frames++;
                     num_2D_frames++;
                     break;

            case  2: 
                     num_palettes++;
                     break;
        }
    }
    return;
}

//############################################################################
void ilda::tell()
{
    cout << "\n\n    total sections  = ";
    cout << num_sections;
    cout << "";
    cout << "\n    total frames    = ";
    cout << num_frames;
    cout << "";
    cout << "\n    total 2D frames = ";
    cout << num_2D_frames;
    cout << "";
    cout << "\n    total 3D frames = ";
    cout << num_3D_frames;
    cout << "";
    cout << "\n    total palettes  = ";
    cout << num_palettes;
    return;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
